@@ -30,15 +30,18 @@
 use std::io;
 use std::io::{Read, Write};
 use std::path::{Path, PathBuf};
+#[cfg(feature = "find")]
 use wildmatch::WildMatch;
 // -- modules
 mod errors;
 mod file;
+pub mod stream;
 mod welcome;
 
 // -- export
 pub use errors::{RemoteError, RemoteErrorType, RemoteResult};
-pub use file::{Directory, Entry, File, Metadata, UnixPex, UnixPexClass};
+pub use file::{File, FileType, Metadata, UnixPex, UnixPexClass};
+pub use stream::{ReadStream, WriteStream};
 pub use welcome::Welcome;
 
 /// Defines the methods which must be implemented in order to setup a Remote file system
@@ -62,10 +65,10 @@ pub trait RemoteFs {
     fn change_dir(&mut self, dir: &Path) -> RemoteResult<PathBuf>;
 
     /// List directory entries at specified `path`
-    fn list_dir(&mut self, path: &Path) -> RemoteResult<Vec<Entry>>;
+    fn list_dir(&mut self, path: &Path) -> RemoteResult<Vec<File>>;
 
     /// Stat file at specified `path` and return Entry
-    fn stat(&mut self, path: &Path) -> RemoteResult<Entry>;
+    fn stat(&mut self, path: &Path) -> RemoteResult<File>;
 
     /// Set metadata for file at specifieed `path`
     fn setstat(&mut self, path: &Path, metadata: Metadata) -> RemoteResult<()>;
@@ -96,21 +99,23 @@ pub trait RemoteFs {
             let path = crate::utils::path::absolutize(&self.pwd()?, path);
             debug!("Removing {}...", path.display());
             let entry = self.stat(path.as_path())?;
-            match entry {
-                Entry::File(_) => self.remove_file(entry.path()),
-                Entry::Directory(d) => {
-                    // list dir
-                    debug!("{} is a directory; removing all directory entries", d.name);
-                    let directory_content = self.list_dir(d.path.as_path())?;
-                    for entry in directory_content.iter() {
-                        self.remove_dir_all(entry.path())?;
-                    }
-                    trace!(
-                        "Removed all files in {}; removing directory",
-                        d.path.display()
-                    );
-                    self.remove_dir(d.path.as_path())
+            if entry.is_dir() {
+                // list dir
+                debug!(
+                    "{} is a directory; removing all directory entries",
+                    entry.name()
+                );
+                let directory_content = self.list_dir(entry.path())?;
+                for entry in directory_content.iter() {
+                    self.remove_dir_all(entry.path())?;
                 }
+                trace!(
+                    "Removed all files in {}; removing directory",
+                    entry.path().display()
+                );
+                self.remove_dir(entry.path())
+            } else {
+                self.remove_file(entry.path())
             }
         } else {
             Err(RemoteError::new(RemoteErrorType::NotConnected))
@@ -140,7 +145,7 @@ pub trait RemoteFs {
     ///
     /// metadata should be the same of the local file.
     /// In some protocols, such as `scp` the `size` field is used to define the transfer size (required by the protocol)
-    fn append(&mut self, path: &Path, metadata: &Metadata) -> RemoteResult<Box<dyn Write>>;
+    fn append(&mut self, path: &Path, metadata: &Metadata) -> RemoteResult<WriteStream>;
 
     /// Create file at path for write.
     /// If the file already exists, its content will be overwritten
@@ -149,10 +154,10 @@ pub trait RemoteFs {
     ///
     /// metadata should be the same of the local file.
     /// In some protocols, such as `scp` the `size` field is used to define the transfer size (required by the protocol)
-    fn create(&mut self, path: &Path, metadata: &Metadata) -> RemoteResult<Box<dyn Write>>;
+    fn create(&mut self, path: &Path, metadata: &Metadata) -> RemoteResult<WriteStream>;
 
     /// Open file at specified path for read.
-    fn open(&mut self, path: &Path) -> RemoteResult<Box<dyn Read>>;
+    fn open(&mut self, path: &Path) -> RemoteResult<ReadStream>;
 
     /// Finalize `create_file` and `append_file` methods.
     /// This method must be implemented only if necessary; in case you don't need it, just return `Ok(())`
@@ -163,7 +168,7 @@ pub trait RemoteFs {
     /// ### Default implementation
     ///
     /// By default this function returns already `Ok(())`
-    fn on_written(&mut self, _writable: Box<dyn Write>) -> RemoteResult<()> {
+    fn on_written(&mut self, _writable: WriteStream) -> RemoteResult<()> {
         Ok(())
     }
 
@@ -176,7 +181,7 @@ pub trait RemoteFs {
     /// ### Default implementation
     ///
     /// By default this function returns already `Ok(())`
-    fn on_read(&mut self, _readable: Box<dyn Read>) -> RemoteResult<()> {
+    fn on_read(&mut self, _readable: ReadStream) -> RemoteResult<()> {
         Ok(())
     }
 
@@ -184,6 +189,7 @@ pub trait RemoteFs {
     /// This method **SHOULD** be implemented **ONLY** when streams are not supported by the current file transfer.
     /// The developer using the client should FIRST try with `create` followed by `on_written`
     /// If the function returns error of kind `UnsupportedFeature`, then he should call this function.
+    /// In case of success, returns the amount of bytes written to the remote file
     ///
     /// ### Default implementation
     ///
@@ -193,14 +199,15 @@ pub trait RemoteFs {
         path: &Path,
         metadata: &Metadata,
         mut reader: Box<dyn Read>,
-    ) -> RemoteResult<()> {
+    ) -> RemoteResult<u64> {
         if self.is_connected() {
             trace!("Opened remote file");
             let mut stream = self.append(path, metadata)?;
             let sz = io::copy(&mut reader, &mut stream)
                 .map_err(|e| RemoteError::new_ex(RemoteErrorType::ProtocolError, e.to_string()))?;
+            self.on_written(stream)?;
             trace!("Written {} bytes to destination", sz);
-            self.on_written(stream)
+            Ok(sz)
         } else {
             Err(RemoteError::new(RemoteErrorType::NotConnected))
         }
@@ -210,6 +217,7 @@ pub trait RemoteFs {
     /// This method SHOULD be implemented ONLY when streams are not supported by the current file transfer.
     /// The developer using the client should FIRST try with `create` followed by `on_written`
     /// If the function returns error of kind `UnsupportedFeature`, then he should call this function.
+    /// In case of success, returns the amount of bytes written to the remote file
     ///
     /// ### Default implementation
     ///
@@ -219,14 +227,15 @@ pub trait RemoteFs {
         path: &Path,
         metadata: &Metadata,
         mut reader: Box<dyn Read>,
-    ) -> RemoteResult<()> {
+    ) -> RemoteResult<u64> {
         if self.is_connected() {
             let mut stream = self.create(path, metadata)?;
             trace!("Opened remote file");
             let sz = io::copy(&mut reader, &mut stream)
                 .map_err(|e| RemoteError::new_ex(RemoteErrorType::ProtocolError, e.to_string()))?;
+            self.on_written(stream)?;
             trace!("Written {} bytes to destination", sz);
-            self.on_written(stream)
+            Ok(sz)
         } else {
             Err(RemoteError::new(RemoteErrorType::NotConnected))
         }
@@ -237,11 +246,12 @@ pub trait RemoteFs {
     /// (since it would work thanks to the default implementation)
     /// The developer using the client should FIRST try with `open` followed by `on_sent`
     /// If the function returns error of kind `UnsupportedFeature`, then he should call this function.
+    /// In case of success, returns the amount of bytes written to the local stream
     ///
     /// ### Default implementation
     ///
     /// By default this function uses the streams function to copy content from reader to writer
-    fn open_file(&mut self, src: &Path, mut dest: Box<dyn Write + Send>) -> RemoteResult<()> {
+    fn open_file(&mut self, src: &Path, mut dest: Box<dyn Write + Send>) -> RemoteResult<u64> {
         if self.is_connected() {
             let mut stream = self.open(src)?;
             trace!("File opened");
@@ -249,7 +259,7 @@ pub trait RemoteFs {
                 .map_err(|e| RemoteError::new_ex(RemoteErrorType::ProtocolError, e.to_string()))?;
             self.on_read(stream)?;
             trace!("Copied {} bytes to destination", sz);
-            Ok(())
+            Ok(sz)
         } else {
             Err(RemoteError::new(RemoteErrorType::NotConnected))
         }
@@ -257,7 +267,8 @@ pub trait RemoteFs {
 
     /// Find files from current directory (in all subdirectories) whose name matches the provided search
     /// Search supports wildcards ('?', '*')
-    fn find(&mut self, search: &str) -> RemoteResult<Vec<Entry>> {
+    #[cfg(feature = "find")]
+    fn find(&mut self, search: &str) -> RemoteResult<Vec<File>> {
         match self.is_connected() {
             true => {
                 // Starting from current directory, iter dir
@@ -276,8 +287,9 @@ pub trait RemoteFs {
     ///
     /// NOTE: DON'T RE-IMPLEMENT THIS FUNCTION, unless the file transfer provides a faster way to do so
     /// NOTE: don't call this method from outside; consider it as private
-    fn iter_search(&mut self, dir: &Path, filter: &WildMatch) -> RemoteResult<Vec<Entry>> {
-        let mut drained: Vec<Entry> = Vec::new();
+    #[cfg(feature = "find")]
+    fn iter_search(&mut self, dir: &Path, filter: &WildMatch) -> RemoteResult<Vec<File>> {
+        let mut drained: Vec<File> = Vec::new();
         // Scan directory
         match self.list_dir(dir) {
             Ok(entries) => {
@@ -287,20 +299,15 @@ pub trait RemoteFs {
                 - if is file: check if it matches `filter`
                     - if it matches `filter`: push to to filter
                 */
-                for entry in entries.iter() {
-                    match entry {
-                        Entry::Directory(dir) => {
-                            // If directory name, matches wildcard, push it to drained
-                            if filter.matches(dir.name.as_str()) {
-                                drained.push(Entry::Directory(dir.clone()));
-                            }
-                            drained.append(&mut self.iter_search(dir.path.as_path(), filter)?);
-                        }
-                        Entry::File(file) => {
-                            if filter.matches(file.name.as_str()) {
-                                drained.push(Entry::File(file.clone()));
-                            }
+                for entry in entries.into_iter() {
+                    if entry.is_dir() {
+                        // If directory name, matches wildcard, push it to drained
+                        if filter.matches(entry.name().as_str()) {
+                            drained.push(entry.clone());
                         }
+                        drained.append(&mut self.iter_search(entry.path(), filter)?);
+                    } else if filter.matches(entry.name().as_str()) {
+                        drained.push(entry);
                     }
                 }
                 Ok(drained)
